/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2021 Slava Aseev <nullptrnine@basealt.ru>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/
#include "kwalletfreedesktopattributes.h"

#include <QDir>
#include <QFile>
#include <QSaveFile>
#include "kwalletd_debug.h"
#include "kwalletfreedesktopcollection.h"
#include "kwalletd.h"

KWalletFreedesktopAttributes::KWalletFreedesktopAttributes(const QString& walletName) {
    QString writeLocation = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    if (!writeLocation.isEmpty() && writeLocation.back() == '5')
        writeLocation.resize(writeLocation.size() - 1);
    _path = writeLocation + '/' + KWalletD::encodeWalletName(walletName) + "_attributes.json";

    read();
}

void KWalletFreedesktopAttributes::read() {
    QByteArray content;
    {
        QFile file(_path);
        file.open(QIODevice::ReadOnly | QIODevice::Text);
        if (!file.isOpen()) {
            qCWarning(KWALLETD_LOG) << "Can't read attributes file " << _path;
            return;
        }
        content = file.readAll();
    }

    auto jsonDoc = QJsonDocument::fromJson(content);
    if (jsonDoc.isObject()) {
        _params = jsonDoc.object();
    }
    else {
        qCWarning(KWALLETD_LOG) << "Can't read attributes: the root element must be an JSON-object: " << _path;
        _params = QJsonObject();
    }
}

void KWalletFreedesktopAttributes::write() const {
    if (_params.empty()) {
        QFile::remove(_path);
        return;
    }

    QSaveFile sf(_path);
    if (!sf.open(QIODevice::WriteOnly | QIODevice::Unbuffered)) {
        qCWarning(KWALLETD_LOG) << "Can't write attributes file: " << _path;
        return;
    }
    sf.setPermissions(QSaveFile::ReadUser | QSaveFile::WriteUser);

    QJsonDocument saveDoc(_params);

    /* TODO: log error */
    sf.write(saveDoc.toJson());
    sf.commit();
}

static QString entryLocationToStr(const EntryLocation& entryLocation) {
    return entryLocation.folder + '/' + entryLocation.key;
}

static EntryLocation splitToEntryLocation(const QString& entryLocation) {
    int slashPos = entryLocation.lastIndexOf('/');
    if (slashPos == -1) {
        // TODO: log error
        return EntryLocation();
    } else {
        return {entryLocation.left(slashPos), entryLocation.right((entryLocation.size() - slashPos) - 1)};
    }
}


void KWalletFreedesktopAttributes::remove(const EntryLocation& entryLocation) {
    _params.remove(entryLocationToStr(entryLocation));
    if (_params.empty())
        QFile::remove(_path);
    else
        write();
}

void KWalletFreedesktopAttributes::deleteFile() {
    QFile::remove(_path);
}

void KWalletFreedesktopAttributes::renameLabel(const EntryLocation& oldLocation, const EntryLocation& newLocation) {
    QString oldLoc = entryLocationToStr(oldLocation);

    auto found = _params.find(oldLoc);
    if (found == _params.end() || !found->isObject()) {
        qCWarning(KWALLETD_LOG) << "Can't rename label (!?)";
        return;
    }
    auto obj = found->toObject();
    _params.erase(found);
    _params.insert(entryLocationToStr(newLocation), obj);

    write();
}

void KWalletFreedesktopAttributes::renameWallet(const QString &newName) {
    QString writeLocation = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    if (!writeLocation.isEmpty() && writeLocation.back() == '5')
        writeLocation.resize(writeLocation.size() - 1);
    QString newPath = writeLocation + '/' + newName + "_attributes.json";

    QFile::rename(_path, newPath);
    _path = newPath;
}

void KWalletFreedesktopAttributes::newItem(const EntryLocation& entryLocation) {
    _params[entryLocationToStr(entryLocation)] = QJsonObject();
}

QList<EntryLocation> KWalletFreedesktopAttributes::matchAttributes(const StrStrMap& attributes) const {
    QList<EntryLocation> items;

    for (auto i = _params.constBegin(); i != _params.constEnd(); ++i) {
        if (!i->isObject())
            continue;

        bool match = true;
        auto itemParams = i->toObject();
        auto foundItemAttribs = itemParams.find("attributes");
        if (foundItemAttribs == itemParams.end() || !foundItemAttribs->isObject())
            continue;
        auto itemAttribs = foundItemAttribs->toObject();

        for (auto i = attributes.constBegin(); i != attributes.constEnd(); ++i) {
            auto foundKey = itemAttribs.find(i.key());
            if (foundKey == itemAttribs.end() || !foundKey->isString() || foundKey->toString() != i.value()) {
                match = false;
                break;
            }
        }

        if (match)
            items += splitToEntryLocation(i.key());
    }

    return items;
}

void KWalletFreedesktopAttributes::setAttributes(const EntryLocation& entryLocation, const StrStrMap& attributes) {
    QJsonObject jsonAttrs;
    for (auto i = attributes.constBegin(); i != attributes.constEnd(); ++i)
        jsonAttrs.insert(i.key(), i.value());

    QString strLocation = entryLocationToStr(entryLocation);

    auto foundParams = _params.find(strLocation);
    QJsonObject params;
    if (foundParams != _params.end() && foundParams->isObject())
        params = foundParams->toObject();
    else
        return;

    if (jsonAttrs.empty())
        params.remove("attributes");
    else
        params["attributes"] = jsonAttrs;

    _params[strLocation] = params;

    write();
}

StrStrMap KWalletFreedesktopAttributes::getAttributes(const EntryLocation& entryLocation) const {
    auto foundObj = _params.find(entryLocationToStr(entryLocation));
    if (foundObj == _params.end() || !foundObj->isObject())
        return StrStrMap();
    auto jsonParams = foundObj->toObject();

    auto foundAttrs = jsonParams.find("attributes");
    if (foundAttrs == jsonParams.end() || !foundAttrs->isObject())
        return StrStrMap();
    auto jsonAttrs = foundAttrs->toObject();

    StrStrMap itemAttrs;

    for (auto i = jsonAttrs.constBegin(); i != jsonAttrs.constEnd(); ++i) {
        if (i.value().isString())
            itemAttrs.insert(i.key(), i.value().toString());
    }

    return itemAttrs;
}

QString KWalletFreedesktopAttributes::getStringParam(const EntryLocation& entryLocation, const QString& paramName, const QString& defaultParam) const {
    auto foundParams = _params.find(entryLocationToStr(entryLocation));
    if (foundParams == _params.end() || !foundParams->isObject())
        return defaultParam;
    auto params = foundParams->toObject();

    auto foundParam = params.find(paramName);
    if (foundParam == params.end() || !foundParam->isString())
        return defaultParam;

    return foundParam->toString();
}

qulonglong KWalletFreedesktopAttributes::getULongLongParam(const EntryLocation& entryLocation, const QString& paramName, qulonglong defaultParam) const {
    auto str = getStringParam(entryLocation, paramName, QString::number(defaultParam));
    bool ok = false;
    auto result = str.toULongLong(&ok);
    return ok ? result : defaultParam;
}


void KWalletFreedesktopAttributes::setParam(const EntryLocation& entryLocation, const QString& paramName, const QString& param) {
    auto entryLoc = entryLocationToStr(entryLocation);
    auto foundParams = _params.find(entryLoc);
    if (foundParams == _params.end() || !foundParams->isObject())
        return;

    auto params = foundParams->toObject();

    params[paramName] = param;
    _params[entryLoc] = params;

    write();
}

void KWalletFreedesktopAttributes::setParam(const EntryLocation& entryLocation, const QString& paramName, qulonglong param) {
    setParam(entryLocation, paramName, QString::number(param));
}

void KWalletFreedesktopAttributes::remove(const FdoUniqueLabel& itemUniqLabel) {
    remove(itemUniqLabel.toEntryLocation());
}

void KWalletFreedesktopAttributes::setAttributes(const FdoUniqueLabel& itemUniqLabel, const StrStrMap& attributes) {
    setAttributes(itemUniqLabel.toEntryLocation(), attributes);
}

StrStrMap KWalletFreedesktopAttributes::getAttributes(const FdoUniqueLabel& itemUniqLabel) const {
    return getAttributes(itemUniqLabel.toEntryLocation());
}


QString KWalletFreedesktopAttributes::getStringParam(const FdoUniqueLabel& itemUniqLabel, const QString& paramName, const QString& defaultParam) const {
    return getStringParam(itemUniqLabel.toEntryLocation(), paramName, defaultParam);
}

qulonglong KWalletFreedesktopAttributes::getULongLongParam(const FdoUniqueLabel& itemUniqLabel, const QString& paramName, qulonglong defaultParam) const {
    return getULongLongParam(itemUniqLabel.toEntryLocation(), paramName, defaultParam);
}

void KWalletFreedesktopAttributes::setParam(const FdoUniqueLabel& itemUniqLabel, const QString& paramName, const QString& param) {
    setParam(itemUniqLabel.toEntryLocation(), paramName, param);
}

void KWalletFreedesktopAttributes::setParam(const FdoUniqueLabel& itemUniqLabel, const QString& paramName, qulonglong param) {
    setParam(itemUniqLabel.toEntryLocation(), paramName, param);
}

QList<EntryLocation> KWalletFreedesktopAttributes::listItems() const {
    QList<EntryLocation> items;
    for (auto i = _params.constBegin(); i != _params.constEnd(); ++i)
        items.push_back(splitToEntryLocation(i.key()));
    return items;
}
