/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2021 Slava Aseev <nullptrnine@basealt.ru>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/
#include "kwalletfreedesktopservice.h"

#include <string.h>
#include "kwalletd.h"
#include "kwalletfreedesktopcollection.h"
#include "kwalletfreedesktopitem.h"
#include "kwalletfreedesktopprompt.h"
#include "kwalletfreedesktopserviceadaptor.h"
#include "kwalletfreedesktopsession.h"
#include <KConfigGroup>
#include <QWidget>
#include <QtCrypto>

[[maybe_unused]] int DBUS_SECRET_SERVICE_META_TYPE_REGISTER = []() {
    qRegisterMetaTypeStreamOperators<StrStrMap>("StrStrMap");
    qDBusRegisterMetaType<StrStrMap>();

    qRegisterMetaTypeStreamOperators<QMap<QString, QString>>("QMap<QString, QString>");
    qDBusRegisterMetaType<QMap<QString, QString>>();

    qDBusRegisterMetaType<FreedesktopSecret>();

    qDBusRegisterMetaType<FreedesktopSecretMap>();

    qDBusRegisterMetaType<PropertiesMap>();

    qDBusRegisterMetaType<QCA::SecureArray>();
    qRegisterMetaTypeStreamOperators<QCA::SecureArray>("QCA::SecureArray");

    return 0;
}();

namespace {
    QString mangleInvalidObjectPathChars(const QString &str) {
        const auto utf8Str = str.toUtf8();
        static constexpr char hex[] = "0123456789abcdef";
        static_assert(sizeof(hex) == 17, "");

        QString mangled;
        mangled.reserve(utf8Str.size());

        for (const auto &c : utf8Str) {
            if ((c < 'A' || c > 'Z') &&
                (c < 'a' || c > 'z') &&
                (c < '0' || c > '9') &&
                c != '_')
            {
                auto cp = static_cast<quint8>(c);
                mangled.push_back('_');
                mangled.push_back(hex[cp >> 4]);
                mangled.push_back(hex[cp & 0x0f]);
            } else {
                mangled.push_back(c);
            }
        }

        return mangled;
    }
}

#define LABEL_NUMBER_PREFIX "__"
#define LABEL_NUMBER_POSTFIX "_"
#define LABEL_NUMBER_REGEX "(^.*)" LABEL_NUMBER_PREFIX "(\\d+)" LABEL_NUMBER_POSTFIX "$"

EntryLocation EntryLocation::fromUniqueLabel(const FdoUniqueLabel& uniqLabel) {
    QString dir;
    QString name = uniqLabel.label;

    int slashPos = uniqLabel.label.lastIndexOf('/');
    if (slashPos == -1) {
        dir = FDO_SECRETS_DEFAULT_DIR;
    } else {
        dir = uniqLabel.label.left(slashPos);
        name = uniqLabel.label.right((uniqLabel.label.size() - dir.size()) - 1);
    }

    return EntryLocation{dir, FdoUniqueLabel::makeName(name, uniqLabel.copyId)};
}

FdoUniqueLabel EntryLocation::toUniqueLabel() const {
    return FdoUniqueLabel::fromEntryLocation(*this);
}

FdoUniqueLabel FdoUniqueLabel::fromEntryLocation(const EntryLocation& entryLocation) {
     auto uniqLabel = FdoUniqueLabel::fromName(entryLocation.key);

    if (entryLocation.folder == FDO_SECRETS_DEFAULT_DIR)
        return uniqLabel;
    else
        return {entryLocation.folder + '/' + uniqLabel.label, uniqLabel.copyId};
}

FdoUniqueLabel FdoUniqueLabel::fromName(const QString& name) {
    static QRegularExpression regexp(LABEL_NUMBER_REGEX);

    auto match = regexp.match(name);
    if (match.hasMatch()) {
        QString strNum = match.captured(2);
        bool ok = false;
        int n = strNum.toInt(&ok);
        if (ok)
            return FdoUniqueLabel{match.captured(1), n};
    }
    return FdoUniqueLabel(name);
}

QString FdoUniqueLabel::makeName(const QString &label, int n) {
    if (n == -1)
        return label;
    else
        return label + LABEL_NUMBER_PREFIX + QString::number(n) + LABEL_NUMBER_POSTFIX;
}

QString FdoUniqueLabel::toName() const {
    return makeName(label, copyId);
}

EntryLocation FdoUniqueLabel::toEntryLocation() const {
    return EntryLocation::fromUniqueLabel(*this);
}

QString KWalletFreedesktopService::wrapToCollectionPath(const QString &itemPath) {
    /* Take only /org/freedesktop/secrets/collection/collection_name */
    return itemPath.section('/', 0, 5);
}

KWalletFreedesktopService::KWalletFreedesktopService(KWalletD *parent)
    : QObject(nullptr), m_parent(parent), m_kwalletrc(QStringLiteral("kwalletrc")) {
    (void)new KWalletFreedesktopServiceAdaptor(this);

    /* register */
    QDBusConnection::sessionBus().registerService(QStringLiteral("org.freedesktop.secrets"));
    QDBusConnection::sessionBus().registerObject(QStringLiteral(FDO_SECRETS_SERVICE_OBJECT), this);

    KConfigGroup walletGroup(&m_kwalletrc, "Wallet");
    if (!parent || !walletGroup.readEntry("Enabled", true))
        return;

    connect(m_parent, static_cast<void (KWalletD::*)(const QString &)>(&KWalletD::walletClosed),
            this, &KWalletFreedesktopService::lockCollection);
    connect(m_parent, &KWalletD::entryUpdated, this, &KWalletFreedesktopService::entryUpdated);
    connect(m_parent, &KWalletD::entryDeleted, this, &KWalletFreedesktopService::entryDeleted);
    connect(m_parent, &KWalletD::entryRenamed, this, &KWalletFreedesktopService::entryRenamed);
    connect(m_parent, &KWalletD::walletDeleted, this, &KWalletFreedesktopService::walletDeleted);
    connect(m_parent, &KWalletD::walletCreated, this, &KWalletFreedesktopService::walletCreated);

    const auto walletNames = backend()->wallets();

    /* Build collections */
    for (const QString &walletName : walletNames) {
        auto objectPath = makeUniqueObjectPath(walletName);
        auto collection = std::unique_ptr<KWalletFreedesktopCollection>(
                new KWalletFreedesktopCollection(this, -1, walletName, objectPath));

        m_collections.emplace(objectPath.path(), std::move(collection));
    }
}

KWalletFreedesktopService::~KWalletFreedesktopService() = default;

QList<QDBusObjectPath> KWalletFreedesktopService::collections() const {
    QList<QDBusObjectPath> result;
    result.reserve(m_collections.size());

    for (const auto &collectionPair : m_collections)
        result.push_back(QDBusObjectPath(collectionPair.first));

    return result;
}

QDBusObjectPath KWalletFreedesktopService::CreateCollection(const QVariantMap &properties,
                                                            const QString &alias,
                                                            QDBusObjectPath &prompt) {
    prompt.setPath("/");

    auto labelIter = properties.find("org.freedesktop.Secret.Collection.Label");
    if (labelIter == properties.end()) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs, "Collection.Label property is missing");
        return QDBusObjectPath("/");
    }
    if (!labelIter->canConvert<QString>()) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                       "Type of Collection.Label property is invalid");
        return QDBusObjectPath("/");
    }

    prompt = nextPromptPath();
    auto fdoPromptPtr = std::unique_ptr<KWalletFreedesktopPrompt>(
            new KWalletFreedesktopPrompt(this, prompt, PromptType::Create, message().service()));
    auto &fdoPrompt = *m_prompts.emplace(prompt.path(), std::move(fdoPromptPtr)).first->second;

    fdoPrompt.appendProperties(labelIter->toString(), QDBusObjectPath("/"), alias);
    fdoPrompt.subscribeForWalletAsyncOpened();

    return QDBusObjectPath("/");
}

FreedesktopSecretMap KWalletFreedesktopService::GetSecrets(const QList<QDBusObjectPath> &items,
                                                           const QDBusObjectPath &session) {
    FreedesktopSecretMap result;

    for (const QDBusObjectPath &itemPath : items) {
        auto item = getItemByObjectPath(itemPath);

        if (item) {
            result.insert(itemPath, item->getSecret(connection(), message(), session));
        } else {
            sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                           "Can't find item at path " + itemPath.path());
            break;
        }
    }

    return result;
}

QList<QDBusObjectPath> KWalletFreedesktopService::Lock(const QList<QDBusObjectPath> &objects,
                                                       QDBusObjectPath &prompt) {
    prompt = QDBusObjectPath("/");
    QList<QDBusObjectPath> result;

    /* Try find in active collections */
    for (const QDBusObjectPath &object : objects) {
        QString collectionPath = wrapToCollectionPath(resolveIfAlias(object.path()));

        auto foundCollection = m_collections.find(collectionPath);
        if (foundCollection != m_collections.end()) {
            int walletHandle = foundCollection->second->walletHandle();
            int rc = m_parent->close(walletHandle, true, FDO_APPID, message());

            if (rc == 0)
                result.push_back(QDBusObjectPath(collectionPath));
            else
                sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                               "Can't lock object at path " + collectionPath);
        } else {
            sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                           "Collection at path " + collectionPath + " does not exist");
        }
    }

    return result;
}

QDBusVariant KWalletFreedesktopService::OpenSession(const QString &algorithm,
                                                    const QDBusVariant &input,
                                                    QDBusObjectPath &result) {
    if (algorithm != "dh-ietf1024-sha256-aes128-cbc-pkcs7") {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                       "Algorithm " + algorithm +
                           " is not supported. (only dh-ietf1024-sha256-aes128-cbc-pkcs7 is supported)");
        return {};
    }

    if (!input.variant().canConvert<QByteArray>()) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                       "Second input argument must be a byte array.");
        return {};
    }

    auto sessionPath = createSession(input.variant().toByteArray());
    result.setPath(sessionPath);

    if (sessionPath != "/")
        return QDBusVariant(
            QVariant(m_sessions[sessionPath]->publicKey().toDH().y().toArray().toByteArray()));
    else
        return QDBusVariant(QVariant(QByteArray()));
}

QDBusObjectPath KWalletFreedesktopService::ReadAlias(const QString &name) {
    QString walletName;

    m_kwalletrc.reparseConfiguration();
    if (name == "default") {
        KConfigGroup cfg(&m_kwalletrc, "Wallet");
        walletName = cfg.readEntry("Default Wallet", QString());
    } else {
        KConfigGroup cfg(&m_kwalletrc, "org.freedesktop.secrets.aliases");
        walletName = cfg.readEntry(name, QString());
    }

    if (!walletName.isEmpty()) {
        auto *collection = getCollectionByWalletName(walletName);
        if (collection)
            return collection->fdoObjectPath();
    }

    return QDBusObjectPath("/");
}

QList<QDBusObjectPath> KWalletFreedesktopService::SearchItems(const StrStrMap &attributes,
                                                              QList<QDBusObjectPath> &locked) {
    QList<QDBusObjectPath> unlocked;

    for (const auto &collectionPair : m_collections) {
        auto &collection = *collectionPair.second;

        if (collection.locked())
            locked += collection.SearchItems(attributes);
        else
            unlocked += collection.SearchItems(attributes);
    }

    return unlocked;
}

void KWalletFreedesktopService::SetAlias(const QString &name, const QDBusObjectPath &collectionPath) {
    auto foundCollection = m_collections.find(collectionPath.path());
    if (foundCollection == m_collections.end())
        return;

    auto *collection = foundCollection->second.get();
    createCollectionAlias(name, collection);
}

QString KWalletFreedesktopService::resolveIfAlias(QString alias) {
    if (alias.startsWith(QStringLiteral(FDO_ALIAS_PATH))) {
        auto path = ReadAlias(alias.remove(0, QStringLiteral(FDO_ALIAS_PATH).size())).path();
        if (path != "/") {
            alias = path;
        } else {
            sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                           "Alias " + alias + " does not exist");
            return {};
        }
    }

    if (!alias.startsWith(FDO_SECRETS_COLLECTION_PATH)) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs, "Collection object path is invalid");
        return {};
    }

    return alias;
}

struct UnlockedObject {
    QString walletName;
    QDBusObjectPath objectPath;
};

QList<QDBusObjectPath> KWalletFreedesktopService::Unlock(const QList<QDBusObjectPath> &objects,
                                                         QDBusObjectPath &prompt) {
    prompt = QDBusObjectPath("/");

    QList<QDBusObjectPath> result;
    QList<UnlockedObject> needUnlock;

    /* Try find in active collections */
    for (const QDBusObjectPath &object : objects) {
        QString strPath = object.path();
        QString collectionPath = wrapToCollectionPath(resolveIfAlias(strPath));

        auto foundCollection = m_collections.find(collectionPath);
        if (foundCollection != m_collections.end()) {
            if (foundCollection->second->locked())
                needUnlock.push_back({foundCollection->second->walletName(), QDBusObjectPath(strPath)});
            else
                result.push_back(QDBusObjectPath(strPath));
        } else {
            sendErrorReply(QDBusError::ErrorType::InvalidObjectPath,
                           "Object " + strPath + " does not exist");
            return {};
        }
    }

    if (!needUnlock.empty()) {
        auto promptPath = nextPromptPath();
        auto fdoPromptPtr = std::unique_ptr<KWalletFreedesktopPrompt>(
                new KWalletFreedesktopPrompt(this, promptPath, PromptType::Open, message().service()));
        auto &fdoPrompt = *m_prompts.emplace(promptPath.path(), std::move(fdoPromptPtr)).first->second;

        prompt = QDBusObjectPath(promptPath);

        for (const auto &toUnlock : asConst(needUnlock))
            fdoPrompt.appendProperties(toUnlock.walletName, toUnlock.objectPath);

        fdoPrompt.subscribeForWalletAsyncOpened();
    }
    return result;
}

QString KWalletFreedesktopService::createSession(const QByteArray &clientKey) {
    if (clientKey.size() < FDO_DH_PUBLIC_KEY_SIZE) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs, "Client public key size is invalid");
        return "/";
    }

    QCA::KeyGenerator keygen;
    auto dlGroup = QCA::DLGroup(keygen.createDLGroup(QCA::IETF_1024));
    if (dlGroup.isNull()) {
        sendErrorReply(QDBusError::ErrorType::InvalidArgs,
                       "createDLGroup failed: maybe libqca-ossl is missing");
        return "/";
    }

    auto privateKey = QCA::PrivateKey(keygen.createDH(dlGroup));
    auto publicKey = QCA::PublicKey(privateKey);
    auto clientPublicKey = QCA::DHPublicKey(dlGroup, QCA::BigInteger(QCA::SecureArray(clientKey)));
    auto commonSecret = privateKey.deriveKey(clientPublicKey);

    auto hmacSalt = QCA::SymmetricKey(QByteArray(FDO_SECRETS_HMAC_SIZE, 0x00));
    auto hmac = QCA::MessageAuthenticationCode("hmac(sha256)", hmacSalt);
    hmac.update(commonSecret.toByteArray());
    auto key = hmac.final();
    hmac.setup(QCA::SymmetricKey(key.toByteArray()));
    hmac.update(QByteArray(1, 0x01));
    auto symmetricKey =
        QCA::SymmetricKey(QByteArray(hmac.final().constData(), FDO_SECRETS_CIPHER_KEY_SIZE));

    QString sessionPath = FDO_SECRETS_SESSION_PATH + QString::number(++m_session_counter);

    auto session = std::unique_ptr<KWalletFreedesktopSession>(
            new KWalletFreedesktopSession(this, publicKey, symmetricKey, sessionPath, connection(), message()));
    m_sessions[sessionPath] = std::move(session);
    return sessionPath;
}

QDBusObjectPath KWalletFreedesktopService::promptUnlockCollection(const QString &walletName, int handle) {
    auto *collection = getCollectionByWalletName(walletName);
    QString objectPath;

    if (collection) {
        collection->onWalletChangeState(handle);
        onCollectionChanged(collection->fdoObjectPath());
        objectPath = collection->fdoObjectPath().path();
    } else {
        auto path = makeUniqueObjectPath(walletName);
        objectPath = path.path();
        auto newCollection = std::unique_ptr<KWalletFreedesktopCollection>(
                new KWalletFreedesktopCollection(this, handle, walletName, path));
        m_collections[objectPath] = std::move(newCollection);
        onCollectionCreated(path);
    }

    return QDBusObjectPath(objectPath);
}

/* Triggered after KWalletD::walletClosed signal */
void KWalletFreedesktopService::lockCollection(const QString &name) {
    auto *collection = getCollectionByWalletName(name);
    if (collection) {
        collection->onWalletChangeState(-1);
        onCollectionChanged(collection->fdoObjectPath());
    }
}

/* Triggered after KWalletD::entryUpdated signal */
void KWalletFreedesktopService::entryUpdated(const QString &walletName, const QString &folder,
                                             const QString &entryName) {
    auto *collection = getCollectionByWalletName(walletName);
    if (!collection)
        return;

    EntryLocation entryLocation{folder, entryName};
    auto *item = collection->findItemByEntryLocation(entryLocation);
    if (item)
        collection->onItemChanged(item->fdoObjectPath());
    else {
        auto objectPath = collection->nextItemPath();
        collection->pushNewItem(entryLocation.toUniqueLabel(), objectPath);
        collection->onItemCreated(objectPath);
    }
}

/* Triggered after KWalletD::entryDeleted signal */
void KWalletFreedesktopService::entryDeleted(const QString &walletName, const QString &folder,
                                             const QString &entryName) {
    auto *collection = getCollectionByWalletName(walletName);
    if (!collection)
        return;

    auto *item = collection->findItemByEntryLocation({folder, entryName});
    if (item)
        collection->onItemDeleted(item->fdoObjectPath());
}

/* Triggered after KWalletD::entryRenamed signal */
void KWalletFreedesktopService::entryRenamed(const QString &walletName, const QString &folder,
                                             const QString &oldName, const QString &newName) {
    auto *collection = getCollectionByWalletName(walletName);
    if (!collection)
        return;

    EntryLocation oldLocation{folder, oldName};
    EntryLocation newLocation{folder, newName};

    auto *item = collection->findItemByEntryLocation(oldLocation);
    if (!item)
        item = collection->findItemByEntryLocation(newLocation);

    if (item) {
        collection->itemAttributes().renameLabel(oldLocation, newLocation);
        item->uniqueLabel(newLocation.toUniqueLabel());
        collection->onItemChanged(item->fdoObjectPath());
    }
}

/* Triggered after KWalletD::walletDeleted signal */
void KWalletFreedesktopService::walletDeleted(const QString &walletName) {
    auto *collection = getCollectionByWalletName(walletName);
    if (collection)
        collection->Delete();
}

/* Triggered after KWalletD::walletCreated signal */
void KWalletFreedesktopService::walletCreated(const QString &walletName) {
    auto objectPath = makeUniqueObjectPath(walletName);
    auto collection = std::unique_ptr<KWalletFreedesktopCollection>(
            new KWalletFreedesktopCollection(this, -1, walletName, objectPath));
    m_collections.emplace(objectPath.path(), std::move(collection));
    onCollectionCreated(objectPath);
}

bool KWalletFreedesktopService::desecret(const QDBusMessage &message, FreedesktopSecret &secret) {
    auto foundSession = m_sessions.find(secret.session.path());

    if (foundSession != m_sessions.end()) {
        KWalletFreedesktopSession &session = *foundSession->second;
        auto decrypted = session.decrypt(message, secret.note, secret.initVector);

        if (decrypted.ok) {
            secret.note = std::move(decrypted.bytes);
            return true;
        }
    }

    return false;
}

bool KWalletFreedesktopService::ensecret(const QDBusMessage &message, FreedesktopSecret &secret) {
    auto foundSession = m_sessions.find(secret.session.path());

    if (foundSession != m_sessions.end()) {
        KWalletFreedesktopSession &session = *foundSession->second;
        auto encrypted = session.encrypt(message, secret.note, secret.initVector);

        if (encrypted.ok) {
            secret.note = std::move(encrypted.bytes);
            return true;
        }
    }

    return false;
}

QDBusObjectPath KWalletFreedesktopService::nextPromptPath() {
    static uint64_t id = 0;
    return QDBusObjectPath(QStringLiteral(FDO_SECRET_SERVICE_PROMPT_PATH) + QStringLiteral("p") +
                           QString::number(id++));
}

QDBusArgument &operator<<(QDBusArgument &arg, const FreedesktopSecret &secret) {
    arg.beginStructure();
    arg << secret.session;
    arg << secret.initVector;
    arg << secret.note;
    arg << secret.mimeType;
    arg.endStructure();
    return arg;
}

const QDBusArgument &operator>>(const QDBusArgument &arg, FreedesktopSecret &secret) {
    arg.beginStructure();
    arg >> secret.session;
    arg >> secret.initVector;
    arg >> secret.note;
    arg >> secret.mimeType;
    arg.endStructure();
    return arg;
}

QDataStream &operator<<(QDataStream &stream, const QCA::SecureArray &value) {
    QByteArray bytes = value.toByteArray();
    stream << bytes;
    explicit_bzero(bytes.data(), bytes.size());
    return stream;
}

QDataStream &operator>>(QDataStream &stream, QCA::SecureArray &value) {
    QByteArray bytes;
    stream >> bytes;
    value = QCA::SecureArray(bytes);
    explicit_bzero(bytes.data(), bytes.size());
    return stream;
}

QDBusArgument &operator<<(QDBusArgument &arg, const QCA::SecureArray &value) {
    QByteArray bytes = value.toByteArray();
    arg << bytes;
    explicit_bzero(bytes.data(), bytes.size());
    return arg;
}

const QDBusArgument &operator>>(const QDBusArgument &arg, QCA::SecureArray &buf) {
    QByteArray byteArray;
    arg >> byteArray;
    buf = QCA::SecureArray(byteArray);
    explicit_bzero(byteArray.data(), byteArray.size());
    return arg;
}

KWalletD *KWalletFreedesktopService::backend() const { return m_parent; }

QDBusObjectPath KWalletFreedesktopService::fdoObjectPath() const {
    return QDBusObjectPath(FDO_SECRETS_SERVICE_OBJECT);
}

KWalletFreedesktopItem *
KWalletFreedesktopService::getItemByObjectPath(const QDBusObjectPath &path) const {
    auto str = path.path();
    if (!str.startsWith(FDO_SECRETS_COLLECTION_PATH))
        return nullptr;

    QString collectionPath = wrapToCollectionPath(str);
    auto collectionPos = m_collections.find(collectionPath);
    if (collectionPos == m_collections.end())
        return nullptr;

    auto &collection = collectionPos->second;
    return collection->getItemByObjectPath(str);
}

KWalletFreedesktopPrompt *
KWalletFreedesktopService::getPromptByObjectPath(const QDBusObjectPath &path) const {
    auto foundPrompt = m_prompts.find(path.path());
    if (foundPrompt != m_prompts.end())
        return foundPrompt->second.get();
    else
        return nullptr;
}

FdoUniqueLabel KWalletFreedesktopService::makeUniqueCollectionLabel(const QString &label) {
    int n = -1;
    auto walletName = label;
    QStringList wallets = backend()->wallets();

    while (wallets.contains(walletName))
        walletName = FdoUniqueLabel::makeName(label, ++n);

    return {label, n};
}

QString KWalletFreedesktopService::makeUniqueWalletName(const QString &labelPrefix) {
    return makeUniqueCollectionLabel(labelPrefix).toName();
}

QDBusObjectPath KWalletFreedesktopService::makeUniqueObjectPath(const QString &walletName) const {
    auto mangled = mangleInvalidObjectPathChars(walletName);
    mangled.insert(0, FDO_SECRETS_COLLECTION_PATH);

    QString result = mangled;
    int postfix = 0;
    while (m_collections.count(result))
        result = mangled + QString::number(postfix++);

    return QDBusObjectPath(result);
}

QStringList KWalletFreedesktopService::readAliasesFor(const QString &walletName) {
    m_kwalletrc.reparseConfiguration();
    KConfigGroup cfg(&m_kwalletrc, "org.freedesktop.secrets.aliases");
    auto map = cfg.entryMap();
    QStringList aliases;

    for (auto i = map.begin(); i != map.end(); ++i)
        if (i.value() == walletName)
            aliases.push_back(i.key());

    KConfigGroup cfgWallet(&m_kwalletrc, "Wallet");
    if (cfgWallet.readEntry("Default Wallet", "") == walletName)
        aliases.push_back("default");

    return aliases;
}

void KWalletFreedesktopService::updateCollectionAlias(const QString &alias,
                                                      const QString &walletName) {
    QString sectName = "org.freedesktop.secrets.aliases";
    QString sectKey = alias;

    if (alias == "default") {
        sectName = "Wallet";
        sectKey = "Default Wallet";
    }

    KConfigGroup cfg(&m_kwalletrc, sectName);
    cfg.writeEntry(sectKey, walletName);
    m_kwalletrc.sync();
}

void KWalletFreedesktopService::createCollectionAlias(const QString &alias,
                                                      const QString &walletName) {
    QString sectName = "org.freedesktop.secrets.aliases";
    QString sectKey = alias;

    if (alias == "default") {
        sectName = "Wallet";
        sectKey = "Default Wallet";
    }

    m_kwalletrc.reparseConfiguration();
    KConfigGroup cfg(&m_kwalletrc, sectName);

    QString prevWalletName = cfg.readEntry(sectKey, QString());
    if (!prevWalletName.isEmpty()) {
        auto *prevCollection = getCollectionByWalletName(prevWalletName);
        if (prevCollection)
            QDBusConnection::sessionBus().unregisterObject(QStringLiteral(FDO_ALIAS_PATH) + alias);
    }

    cfg.writeEntry(sectKey, walletName);
    m_kwalletrc.sync();

    auto *collection = getCollectionByWalletName(walletName);
    if (collection)
        QDBusConnection::sessionBus().registerObject(QStringLiteral(FDO_ALIAS_PATH) + alias,
                                                     collection);
}

void KWalletFreedesktopService::createCollectionAlias(const QString &alias,
                                            KWalletFreedesktopCollection *collection) {
    QString sectName = "org.freedesktop.secrets.aliases";
    QString sectKey = alias;

    if (alias == "default") {
        sectName = "Wallet";
        sectKey = "Default Wallet";
    }

    m_kwalletrc.reparseConfiguration();
    KConfigGroup cfg(&m_kwalletrc, sectName);

    QString prevWalletName = cfg.readEntry(sectKey, "");
    if (!prevWalletName.isEmpty()) {
        auto *prevCollection = getCollectionByWalletName(prevWalletName);
        if (prevCollection)
            QDBusConnection::sessionBus().unregisterObject(QStringLiteral(FDO_ALIAS_PATH) + alias);
    }

    cfg.writeEntry(sectKey, collection->walletName());
    m_kwalletrc.sync();
    QDBusConnection::sessionBus().registerObject(QStringLiteral(FDO_ALIAS_PATH) + alias,
                                                 collection);
}

void KWalletFreedesktopService::removeAlias(const QString &alias) {
    if (alias == "default")
        return;

    KConfigGroup cfg(&m_kwalletrc, "org.freedesktop.secrets.aliases");
    cfg.deleteEntry(alias);
    m_kwalletrc.sync();
    QDBusConnection::sessionBus().unregisterObject(QStringLiteral(FDO_ALIAS_PATH) + alias);
}

KWalletFreedesktopCollection *
KWalletFreedesktopService::getCollectionByWalletName(const QString &walletName) const {
    for (const auto &collectionKeyValue : m_collections) {
        auto collection = collectionKeyValue.second.get();
        if (collection->walletName() == walletName)
            return collection;
    }

    return nullptr;
}

void KWalletFreedesktopService::deletePrompt(const QString &objectPath) {
    auto foundPrompt = m_prompts.find(objectPath);
    if (foundPrompt == m_prompts.end())
        return;

    /* This can be called in the context of the prompt that is currently being
     * deleted. Therefore, we should schedule deletion on the next event loop iteration
     */
    foundPrompt->second->deleteLater();
    foundPrompt->second.release();
    m_prompts.erase(foundPrompt);
}

void KWalletFreedesktopService::deleteSession(const QString &objectPath) {
    auto foundSession = m_sessions.find(objectPath);
    if (foundSession == m_sessions.end())
        return;

    /* This can be called in the context of the session that is currently being
     * deleted. Therefore, we should schedule deletion on the next event loop iteration
     */
    foundSession->second->deleteLater();
    foundSession->second.release();
    m_sessions.erase(foundSession);
}

void KWalletFreedesktopService::onCollectionCreated(const QDBusObjectPath &path) {
    Q_EMIT CollectionCreated(path);

    QVariantMap props;
    props.insert("Collections", QVariant::fromValue(collections()));
    onPropertiesChanged(props);
}

void KWalletFreedesktopService::onCollectionChanged(const QDBusObjectPath &path) {
    Q_EMIT CollectionChanged(path);
}

void KWalletFreedesktopService::onCollectionDeleted(const QDBusObjectPath &path) {
    auto collectionMapPos = m_collections.find(path.path());
    if (collectionMapPos == m_collections.end())
        return;
    auto &collectionPair = *collectionMapPos;
    collectionPair.second->itemAttributes().deleteFile();

    /* This can be called in the context of the collection that is currently being
     * deleted. Therefore, we should schedule deletion on the next event loop iteration
     */
    collectionPair.second->deleteLater();
    collectionPair.second.release();
    m_collections.erase(collectionMapPos);

    Q_EMIT CollectionDeleted(path);

    QVariantMap props;
    props["Collections"] = QVariant::fromValue(collections());
    onPropertiesChanged(props);
}

void KWalletFreedesktopService::onPropertiesChanged(const QVariantMap &properties) {
    auto msg = QDBusMessage::createSignal(fdoObjectPath().path(),
                                          QStringLiteral("org.freedesktop.DBus.Properties"),
                                          QStringLiteral("PropertiesChanged"));
    auto args = QVariantList();
    args << "org.freedesktop.Secret.Service" << properties << QStringList();
    msg.setArguments(args);
    QDBusConnection::sessionBus().send(msg);
}

QDataStream &operator<<(QDataStream &stream, const QDBusObjectPath &value) {
    return stream << value.path();
}

QDataStream &operator>>(QDataStream &stream, QDBusObjectPath &value) {
    QString str;
    stream >> str;
    value = QDBusObjectPath(str);
    return stream;
}

const QDBusArgument &operator>>(const QDBusArgument &arg, PropertiesMap &value) {
    arg.beginMap();
    value.map.clear();

    while (!arg.atEnd()) {
        arg.beginMapEntry();
        QString key;
        QVariant val;
        arg >> key >> val;

        /* For org.freedesktop.Secret.Item.Attributes */
        if (val.canConvert<QDBusArgument>()) {
            auto metaArg = val.value<QDBusArgument>();
            StrStrMap metaMap;
            metaArg >> metaMap;
            val = QVariant::fromValue(metaMap);
        }
        value.map.insert(key, val);

        arg.endMapEntry();
    }
    arg.endMap();

    return arg;
}

QDBusArgument &operator<<(QDBusArgument &arg, const PropertiesMap &value) {
    arg << value.map;
    return arg;
}
